#include <stdint.h>


using regptr_t = uint32_t;
using reg32_t  = uint32_t;
using reg16_t  = uint16_t;

/****************************************************************************+**/

/*!< Peripheral memory map */
static constexpr regptr_t  peripheral_base 	= 0x40000000UL; 
static constexpr regptr_t  apb1_base  		=  peripheral_base;
static constexpr regptr_t  apb2_base 		=  peripheral_base + 0x00010000UL;
static constexpr regptr_t  ahb1_base 		=  peripheral_base + 0x00020000UL;
static constexpr regptr_t  ahb2_base 		=  peripheral_base + 0x10000000UL;

/*!< APB1 peripherals */
enum class apb1_t {
	tim2_base   	=	(apb1_base + 0x0000UL),
	tim3_base 		=	(apb1_base + 0x0400ul),
	tim4_base       =   (apb1_base + 0x0800ul),
	tim5_base       =   (apb1_base + 0x0c00ul),
	rtc_base        =   (apb1_base + 0x2800ul),
	wwdg_base       =   (apb1_base + 0x2c00ul),
	iwdg_base       =   (apb1_base + 0x3000ul),
	i2s2ext_base    =   (apb1_base + 0x3400ul),
	spi2_base       =   (apb1_base + 0x3800ul),
	spi3_base       =   (apb1_base + 0x3c00ul),
	i2s3ext_base    =   (apb1_base + 0x4000ul),
	usart2_base     =   (apb1_base + 0x4400ul),
	i2c1_base       =   (apb1_base + 0x5400ul),
	i2c2_base       =   (apb1_base + 0x5800ul),
	i2c3_base       =   (apb1_base + 0x5c00ul),
	pwr_base        =   (apb1_base + 0x7000ul),
};


/*!< APB2 peripherals */
enum class apb2_t {
	tim1_base         =    (apb2_base + 0x0000ul),
	usart1_base       =    (apb2_base + 0x1000ul),
	usart6_base       =    (apb2_base + 0x1400ul),
	adc1_base         =    (apb2_base + 0x2000ul),
	adc1_common_base  =    (apb2_base + 0x2300ul),
	sdio_base         =    (apb2_base + 0x2c00ul),
	spi1_base         =    (apb2_base + 0x3000ul),
	spi4_base         =    (apb2_base + 0x3400ul),
	syscfg_base       =    (apb2_base + 0x3800ul),
	exti_base         =    (apb2_base + 0x3c00ul),
	tim9_base         =    (apb2_base + 0x4000ul),
	tim10_base        =    (apb2_base + 0x4400ul),
	tim11_base        =    (apb2_base + 0x4800ul),
	spi5_base         =    (apb2_base + 0x5000ul),
};


/*!< AHB1 peripherals */
enum class ahb1_t {
	gpioa_base        =    (ahb1_base + 0x0000ul),
	gpiob_base        =    (ahb1_base + 0x0400ul),
	gpioc_base        =    (ahb1_base + 0x0800ul),
	gpiod_base        =    (ahb1_base + 0x0c00ul),
	gpioe_base        =    (ahb1_base + 0x1000ul),
	gpioh_base        =    (ahb1_base + 0x1c00ul),
	crc_base          =    (ahb1_base + 0x3000ul),
	rcc_base          =    (ahb1_base + 0x3800ul),
	flash_r_base      =    (ahb1_base + 0x3c00ul),
	dma1_base         =    (ahb1_base + 0x6000ul),
	dma1_stream0_base =    (dma1_base + 0x010ul),
	dma1_stream1_base =    (dma1_base + 0x028ul),
	dma1_stream2_base =    (dma1_base + 0x040ul),
	dma1_stream3_base =    (dma1_base + 0x058ul),
	dma1_stream4_base =    (dma1_base + 0x070ul),
	dma1_stream5_base =    (dma1_base + 0x088ul),
	dma1_stream6_base =    (dma1_base + 0x0a0ul),
	dma1_stream7_base =    (dma1_base + 0x0b8ul),
	dma2_base         =    (ahb1_base + 0x6400ul),
	dma2_stream0_base =    (dma2_base + 0x010ul),
	dma2_stream1_base =    (dma2_base + 0x028ul),
	dma2_stream2_base =    (dma2_base + 0x040ul),
	dma2_stream3_base =    (dma2_base + 0x058ul),
	dma2_stream4_base =    (dma2_base + 0x070ul),
	dma2_stream5_base =    (dma2_base + 0x088ul),
	dma2_stream6_base =    (dma2_base + 0x0a0ul),
	dma2_stream7_base =    (dma2_base + 0x0b8ul),
};

enum class tim_base_t {
	tim1_base       =   static_cast<regptr_t>(apb2_t::tim1_base),	
	tim2_base   	=	static_cast<regptr_t>(apb1_t::tim2_base),
	tim3_base 		=	static_cast<regptr_t>(apb1_t::tim3_base),
	tim4_base       =   static_cast<regptr_t>(apb1_t::tim4_base),
	tim5_base       =   static_cast<regptr_t>(apb1_t::tim5_base),
	tim9_base       =   static_cast<regptr_t>(apb2_t::tim9_base),
	tim10_base      =   static_cast<regptr_t>(apb2_t::tim10_base),
	tim11_base      =   static_cast<regptr_t>(apb2_t::tim11_base),
};

enum class gpio_base_t {
	gpioa_base     =    static_cast<regptr_t>(ahb1_t::gpioa_base),
	gpiob_base     =    static_cast<regptr_t>(ahb1_t::gpiob_base),
	gpioc_base     =    static_cast<regptr_t>(ahb1_t::gpioc_base),
	gpiod_base     =    static_cast<regptr_t>(ahb1_t::gpiod_base),
	gpioe_base     =    static_cast<regptr_t>(ahb1_t::gpioe_base),
	gpioh_base     =    static_cast<regptr_t>(ahb1_t::gpioh_base),
};

enum class tim_reg_offset_t {
	CR1		= 0x00,		// TIM control register 1
	CR2     = 0x04,		// TIM control register 2
	SMCR    = 0x08,		// TIM slave mode control register 
	DIER	= 0x0C,		// DMA interrupt enable register
	SR		= 0x10,		// status register
	EGR		= 0x14,		// event generation register
	CCMR1	= 0x18,		// capture/compare mode register 1
	CCMR2	= 0x1C,		// capture/compare mode register 2
	CCER	= 0x20,		// capture/compare enable register
	CNT		= 0x24,		// counter register
	PSC		= 0x28,		// prescaler
	ARR		= 0x2C,		// auto-reload register
	RCR		= 0x30,		// repetition counter register
	CCR1	= 0x34,		// capture/compare register 1
	CCR2	= 0x38,		// capture/compare register 2
	CCR3	= 0x3C,		// capture/compare register 3
	CCR4	= 0x40,		// capture/compare register 4
	BDTR	= 0x44,		// break and dead-time register
	DCR		= 0x48,		// DMA control register
	DMAR	= 0x4C,		// DMA address for full transfer
	OR		= 0x50, 	// option register
};

// counter features
enum class count_direction_t {
    count_up,
    count_down,
    count_up_down,
};

enum class update_event_t {
    enable,
    disable,
};

// timer features 
struct tim_counter_t{

    static constexpr void start() {};
    static constexpr void stop() {};

    template <typename T, typename ... Types>
    void static constexpr config(T v1, Types ... v2){
        cfg(v1);
        cfg(v2 ...);
    };

private:

    void cfg(){
        // ignore  
    }; 

    static constexpr void cfg(count_direction_t cd){};
    static constexpr void cfg(update_event_t ue){};

 }; 



template<tim_base_t b>
struct timer_t{
    static constexpr regptr_t base = b;
    static constexpr void enable(){

    };
    static constexpr void disable(){}; 
    tim_counter_t counter;
 

};

using TIM1 = timer_t<tim_base_t::tim1_base>;


int main() {

    TIM1 myTimer;
    myTimer.enable();
    
    myTimer.counter.config(
        count_direction_t::count_up,
        update_event_t::enable
    );
    
    
    myTimer.counter.start();
    myTimer.counter.stop();


}

